definition module Geometry

/**	This module defines the part of the SoccerFun API that is concerned with the footballer dimensions.
*/
import StdClass

::	Position
	=	{	px		:: !XPos		// x-coordinate in plane (0.0<=px)
		,	py		:: !YPos		// y-coordinate in plane (0.0<=py)
		}
::	Position3D
	=	{	pxy		:: !Position
		,	pz		:: !ZPos
		}
:: XPos				:== Metre
:: YPos				:== Metre
:: ZPos				:== Metre
:: Metre			:== Real		

//  position3D is within the cube around #param5 with measures of #param2t/m4		
inRadiusOfPosition	:: !Position3D !XRadius YRadius ZRadius !Position -> Bool
:: XRadius			:== Metre
:: YRadius			:== Metre
:: ZRadius			:== Metre

angleHowFarFromPi	:: !Angle -> Angle
angleHowFarFromAngle:: !Angle !Angle -> Angle

instance zero Position
instance zero Position3D
instance +    RVector
instance +    RVector3D
instance -    RVector
instance -    RVector3D
instance zero RVector
instance zero RVector3D
instance ~    RVector
instance ~    RVector3D
instance zero Speed
instance zero Speed3D

/** move_point v p
		moves point p over vector v.
*/
move_point				:: !RVector   !Position   -> Position
move_point3D			:: !RVector3D !Position3D -> Position3D

/**	point_to_rectangle (a,b) c
		returns c if (point_in_rectangle (a,b) c) and the projected point c` of c that is exactly
		on the closest edge of rectangle (a,b).
*/
point_to_rectangle		:: !(!Position,!Position) !Position -> Position
/** point_in_rectangle (a,b) c
		returns True iff point c is inside the rectangle determined by 
		the diagonal corner points a and b.
*/
point_in_rectangle		:: !(!Position,!Position) !Position -> Bool


//	same as above but now a cilinder instead of a bar
inCircleRadiusOfPosition:: !Position3D !XRadius ZRadius !Position -> Bool
		
::	Speed												// speed of an object
	=	{	direction	:: !Angle						// direction of object
		,	velocity	:: !Velocity					// velocity of object
		}
::	Speed3D												// speed of an object in space
	=	{	vxy			:: !Speed						// surface speed of object
		,	vz			:: !Velocity					// velocity in z-axis (positive: goes up; negative: goes down; zero: horizontally)
		}
::	Velocity			:== Real						// velocity in metre/second		

::	Angle				:== Radian						// angle in radians, clockwise
::	Radian				:== Real

::	RVector
	=	{	dx			:: !Metre						// difference in x-coordinate (|dx| <= 1.0)
		,	dy			:: !Metre						// difference in y-coordinate (|dy| <= 1.0)
		}
::	RVector3D
	=	{	dxy			:: !RVector
		,	dz			:: !Metre
		}

instance ==          Speed
instance ==          Speed3D
instance toString    Speed
instance toString    Speed3D

class    toSpeed     a :: !a -> Speed
class    fromSpeed   a :: !Speed -> a
class    toSpeed3D   a :: !a -> Speed3D
class    fromSpeed3D a :: !Speed3D -> a
instance toSpeed     Speed3D
instance fromSpeed   Speed3D
instance toSpeed3D   Speed
instance fromSpeed3D Speed

pi :==  3.1415926535897932384

/**	scale_vector k {dx,dy}
		returns {k*dx,k*dy}
	scale_vector3D k {dxy,dz}
		returns {scale_vector k dxy,k*dz}
*/
scale_vector   :: !Real !RVector   -> RVector
scale_vector3D :: !Real !RVector3D -> RVector3D

/**	between_points (a,b) c
		returns True iff point c is on the line between point a and point b.
*/
between_points :: !(!Position,!Position) !Position -> Bool

class toRVector   a :: !a -> RVector
instance toRVector   Real				// interpret Real as angle in radians
instance toRVector   Position

class dist a b :: !a !b -> Real
instance dist Real       Real
instance dist Position   Position
instance dist Position3D Position3D
instance dist Position   Position3D
instance dist Position3D Position

class toPosition     a :: !a -> Position
class fromPosition   a :: !Position -> a
class toPosition3D   a :: !a -> Position3D
class fromPosition3D a :: !Position3D -> a
instance toPosition     (!Real,!Real)
instance toPosition     Position3D
instance fromPosition   (!Real,!Real)
instance fromPosition   Position3D
instance toPosition3D   (!Real,!Real,!Real)
instance toPosition3D   Position
instance fromPosition3D (!Real,!Real,!Real)
instance fromPosition3D Position

instance == Position
instance == Position3D
instance toString Position
instance toString Position3D

/** size_vector {dx,dy} = sqrt (dx^2 + dy^2)
	size_vector3D {dxy,dz} = sqrt (dxy.dx^2 + dxy.dy^2 + dz^2)
*/
size_vector				:: !RVector   -> Real
size_vector3D			:: !RVector3D -> Real

/** orthogonal a
		returns the left- and right- orthogonal angles to a
*/
orthogonal				:: !Angle -> (!Angle,!Angle)
oppositeAngle			:: !Angle -> Angle

//use for Kickball and Headball
angleWithObject			::   !Position         !Position -> Angle
//use for Run and Rotate
angleWithObjectForRun	:: !(!Position,!Angle) !Position -> Angle 

/**	Conversion of radians to degrees and vice versa:
*/
::	Degrees				:== Int		// 0 <= degree < 360 (clockwise)

fromRadian				:: !Radian  -> Degrees
toRadian				:: !Degrees -> Radian
