implementation module matchControl

import StdEnvExt
import Gui2D				// we choose the 2D GUI version of SoccerFun
import Referee

doSoccerFun :: !*World -> *World
doSoccerFun world					= SoccerFunGUI2D world

getMatchStart :: !Team !Team !FootballField !Referee !PlayingTime !*env -> (!Match,!*env) | TimeEnv, FileSystem env
getMatchStart fstTeam sndTeam field referee time env
	# (seed,env)					= getNewRandomSeed env
	= ( { team1						= validateTeam fstTeam
		, team2						= validateTeam sndTeam
		, theBall					= Free (ballAtCenter field)
		, theField					= field
		, theReferee				= referee
		, playingHalf				= FirstHalf
		, playingTime				= time
		, unittime					= 0.1
		, score						= (0,0)
		, nextRandomP				= nextRandomP
		, seed						= seed
		}
	  , env)


showMinutes :: !Minutes -> String
showMinutes m						= toString (entier m) <+++ ":" <+++ entier ((m - (toReal (entier m)))*60.0)

stepMatch :: !Match -> (!(![RefereeAction],![FootballerWithAction]),!Match)
stepMatch match
	# (refereeActions,  match)		= refereeTurn match
	# match							= performRefereeActions refereeActions match
	# (intendedActions, match)		= playersTurn   refereeActions  match
	# (succeededActions,match)		= selectActions intendedActions match
	# match							= performFootballerActions intendedActions succeededActions match
	# match							= advanceTime match
	= ((refereeActions,succeededActions),match)
where
/*	playersTurn match
		lets every footballer player conjure an initiative.
*/	playersTurn :: ![RefereeAction] !Match -> (![FootballerWithAction],!Match)
	playersTurn refereeActions match= (intendedActions,new_match)
	where
		actionsOfTeam1				= map (think refereeActions match.theBall match.team2) (singleOutElems match.team1)
		actionsOfTeam2				= map (think refereeActions match.theBall match.team1) (singleOutElems match.team2)
		new_match					= {match & team1 = map snd actionsOfTeam1,team2 = map snd actionsOfTeam2}
		intendedActions				= [(action,playerID) \\ (action,{playerID}) <- actionsOfTeam1 ++ actionsOfTeam2]
		
		think :: ![RefereeAction] !FootballState ![Footballer] !(!Footballer,![Footballer]) -> (!FootballerAction,!Footballer)
		think refereeActions ballstate opponents (me=:{Footballer | effect,brain=brain=:{ai,m}},ownTeam)
			| isNothing effect		= (action,new_me)
			| otherwise				= checkIfNotOnGround (fromJust effect) action new_me
		where
			(action,new_m)			= ai {referee=refereeActions,football=ballstate,others=ownTeam ++ opponents,me=me,memory=m}
			new_me					= {Footballer | me & brain = {Brain | brain & m=new_m}}
			
			checkIfNotOnGround :: !FootballerEffect !FootballerAction !Footballer -> (!FootballerAction,!Footballer)
			checkIfNotOnGround (OnTheGround i) action fb
				| i <= 0			= (action,              {fb & effect = Nothing})
				| otherwise			= (allowOnlyPlayTheater,{fb & effect = Just (OnTheGround (i-1))
									                            , stamina= alter_stamina ballstate fb zero zero})
			where
				allowOnlyPlayTheater= case action of
										PlayTheater	= action
										_			= Move {zero & direction = fb.speed.direction} zero //Run {zero & direction = fb.speed.direction}
			checkIfNotOnGround _ action fb
									= (action,fb)
		
/*	selectActions actions match
		removes all failing actions, and returns the list of remaining succeeding actions.
		It updates the random stream in match, and neutralizes actions of tackled players.
*/	selectActions :: ![FootballerWithAction] !Match -> (![FootballerWithAction],!Match)
	selectActions actions match=:{seed}
		# (seed,succeededActions)				= validActions match seed actions
		# ((successTackles,failedTackles),seed)	= analyseTackleActions match succeededActions seed
		# succeededActions						= filter (isNoTackleVictim successTackles) (succeededActions except failedTackles)
		= (succeededActions,{match & seed=seed})
	where
	/*	actions that are always valid: {Move, Run, Rotate, Feint, Schwalbe, PlayTheater}
		actions that may have success: {Tackle} (looked at later at performactions)
		actions where at most one can succeed: {GainBall, KickBall, HeadBall, Catch}
	*/	validActions :: !Match !RandomSeed ![FootballerWithAction] -> (!RandomSeed,![FootballerWithAction])
		validActions match seed []			= (seed,[])
		validActions match seed actions
			| isJust ball_action			= (seed1,[fromJust ball_action:other_actions])
			| otherwise						= (seed1,other_actions)
		where
			allFootballers					= match.team1 ++ match.team2
			(ball_actions,other_actions)	= spanfilter (isActionOnBall o fst) actions
			(seed1,ball_action)				= selectBallAction match.nextRandomP match.theBall allFootballers seed ball_actions
			
			selectBallAction :: !(St RandomSeed P) !FootballState ![Footballer] !RandomSeed ![FootballerWithAction] -> (!RandomSeed,!Maybe FootballerWithAction)
			selectBallAction nextRandomP ballstate allFootballers seed desiredActions
				# (ps,seed)					= iterateStn (length desiredActions) nextRandomP seed
				# odds						= [ (successOfAction ballstate allFootballers action (if (p==one) p (makeRandomRealistic p)),action)
											  \\ action <- desiredActions
											   & p      <- ps
											  ]
				# ok_odds					= filter (\(p,a) -> p > zero) odds
				| isEmpty ok_odds			= (seed,Nothing)
				# max_odd_prob				= maxList (map fst ok_odds)
				# bestActions				= [a \\ (p,a) <- ok_odds | p >= max_odd_prob]
				# (p,seed)					= nextRandomP seed
				| p==one					= (seed,Just (hd bestActions))
				| otherwise					= (seed,Just (bestActions !! (entier (p * toReal (length bestActions)))))
			where
				successOfAction :: !FootballState ![Footballer] !FootballerWithAction !P -> P
				successOfAction ballstate allPlayers (action,who) p
											= myFatigue * myHealth * p * success_of_action
				where
					success_of_action		= if (isGainBall  action && ballGainable  && ballAtGainSpeed)  success_gaining
											 (if (isCatchBall action && ballCatchable && ballAtCatchSpeed) success_catching
											 (if (isKickBall  action && ballKickable)                      success_kicking
											 (if (isHeadBall  action && ballHeadable)                      success_heading
											 	                                                           zero
											 )))
					me						= find1 (identify_player who) allPlayers
					myFatigue 				= me.stamina
					myHealth				= me.health
					mySkills 				= skillsAsList me
					myLength 				= me.length
					iGainWell				= isMember Gaining  mySkills
					iKickWell				= isMember Kicking  mySkills
					iHeadWell				= isMember Heading  mySkills
					iCatchWell				= isMember Catching mySkills
					
					ballGainable			= d_player_ball <= maxGainReach  me
					ballKickable			= d_player_ball <= maxKickReach  me
					ballHeadable			= d_player_ball <= maxHeadReach  me
					ballCatchable			= d_player_ball <= maxCatchReach me
					ballAtGainSpeed			= d_velocity    <= maxGainVelocityDifference  me d_player_ball
					ballAtCatchSpeed		= d_velocity    <= maxCatchVelocityDifference me d_player_ball
					d_speed 				= {zero & dxy = scale_vector me.speed.velocity (toRVector me.speed.direction)}
												-
											  {       dxy = scale_vector theBall.ballSpeed.vxy.velocity (toRVector theBall.ballSpeed.vxy.direction)
											  ,       dz  = theBall.ballSpeed.vz
											  }
					d_velocity 				= size_vector3D d_speed
					
					theBall 				= getFootball ballstate allPlayers
					d_player_ball 			= dist (toPosition3D me.pos) theBall.ballPos
					
					others_with_ball		= [fb \\ fb <- allPlayers | ballIsGainedBy fb.playerID ballstate && not (identify_player who fb)]
					other_has_ball			= not (isEmpty others_with_ball)
					otherDribblesWell		= isMember Dribbling (skillsAsList (hd others_with_ball))
					
					success_gaining			= if (ballIsFree ballstate) (lengthPenalty * if iGainWell 0.95 0.8)
											 (if  other_has_ball        (lengthPenalty * if iGainWell 0.75 0.3 * if otherDribblesWell 0.6 1.0)
												                        1.0)
					success_kicking			= if (ballIsFree ballstate) (lengthBonus * if iKickWell 0.95 0.85)
											 (if  other_has_ball        (lengthBonus * if iKickWell 0.80 0.70 * if otherDribblesWell 0.7 1.0)
												                        1.0)
					success_heading			= if iHeadWell  0.95 0.9
					success_catching		= if iCatchWell 1.0  0.95
					lengthBonus				= (myLength-1.2) ^ 0.15
					lengthPenalty			= (2.6-myLength) ^ 0.1
		
		analyseTackleActions :: !Match ![FootballerWithAction] !RandomSeed -> (!(![FootballerWithAction],![FootballerWithAction]),!RandomSeed)
		analyseTackleActions match performedActions seed
			= spanfilterSt (isPossibleTackle match) [action \\ action <- performedActions | isFootballerTackle (fst action)] seed
		where
			isPossibleTackle :: !Match !FootballerWithAction !RandomSeed -> (!Bool,!RandomSeed)
			isPossibleTackle match=:{team1,team2} (Tackle victimID _,playerID) seed
				| d_me_victim > maxTackleReach offender								// victim is out of reach
								= (False,seed)
				# (p,seed)		= nextRandomP seed
				| otherwise		= (avg [p,chanceOfSuccess] > 0.5,seed)				// victim is within reach, but tackle may fail
			where
				allPlayers		= team1 ++ team2
				offender		= find1 (identify_player playerID) allPlayers
				victim			= find1 (identify_player victimID) allPlayers
				d_me_victim		= dist offender.pos victim.pos
				chanceOfSuccess	= avg [1.0 - d_me_victim, if (isMember Tackling (skillsAsList offender)) 0.9 0.7]
		
		isNoTackleVictim :: ![FootballerWithAction] !FootballerWithAction -> Bool
		isNoTackleVictim tackles (action,playerID)
								= isSchwalbe    action 
									|| 
								  isPlayTheater action 
								  	|| 
								  isEmpty [victim \\ (Tackle victim _,_) <- tackles | victim==playerID]

/*	refereeTurn match
		determines whether the rules of soccer are adhered to and yields a list of referee actions.
*/	refereeTurn :: !Match -> (![RefereeAction],!Match)
	refereeTurn match=:{theReferee=referee=:{Referee | brain=brain=:{ai,m}},theBall,playingHalf,team1,team2,playingTime,unittime,seed}
											= (refereeActions,{match & theReferee=new_referee,seed=new_seed})
	where
		(refereeActions,(new_m,new_seed))	= ai playingTime unittime theBall playingHalf team1 team2 (m,seed)
		new_referee							= {Referee | referee & brain={Brain | brain & m=new_m}}
		
/*	performRefereeActions refereeActions match
		performs for each football player in match his succeededAction, informs them about the referee actions, and moves the ball. 
*/	performRefereeActions :: ![RefereeAction] !Match -> Match
	performRefereeActions refActions match	= foldl doRefereeEvent match refActions
	where
		doRefereeEvent :: !Match !RefereeAction -> Match
		doRefereeEvent theMatch=:{playingHalf,theField,team1,team2} refereeAction
			| isAlterMatchBallAndTeams		= {theMatch & theBall=Free (mkFootball pos zero)}
			| isGameProgressEvent			= gameProgress theMatch
			| isDisplaceTeamsEvent			= {theMatch & team1=map (displacePlayer ds) team1,team2=map (displacePlayer ds) team2}
			| isReprimandEvent				= let (team1`,team2`)		= reprimandPlayer (nameOf team1) tef repr (team1,team2) in {theMatch & team1=team1`,team2=team2`}
			| otherwise						= theMatch
		where
			(isAlterMatchBallAndTeams,pos)	= case refereeAction of
												DirectFreeKick _ pos	= (True,pos)
												ThrowIn        _ pos	= (True,pos)
												Corner         _ _		= (True,fromJust (getKickPos theField playingHalf refereeAction))
												GoalKick       _		= (True,fromJust (getKickPos theField playingHalf refereeAction))
												Penalty        _		= (True,fromJust (getKickPos theField playingHalf refereeAction))
												CenterKick     _		= (True,fromJust (getKickPos theField playingHalf refereeAction))
												otherwise				= (False,undef)
			(isGameProgressEvent,gameProgress)
											= case refereeAction of
												GameOver				= (True,\theMatch -> {theMatch & playingTime=zero})
												AddTime t				= (True,\theMatch -> {theMatch & playingTime=theMatch.playingTime+t})
												EndHalf					= (True,\theMatch -> {theMatch & playingHalf=SecondHalf})
												Goal    t				= (True,\theMatch=:{score=(w,e)} -> {theMatch & score=if (t==Team1) (w+1,e) (w,e+1)})
												otherwise				= (False,undef)
			(isDisplaceTeamsEvent,ds)		= case refereeAction of
												DisplacePlayers ds		= (True, ds)
												otherwise				= (False,undef)
			(isReprimandEvent,(tef,repr))	= case refereeAction of
												ReprimandPlayer p r		= (True, (p,r))
												otherwise				= (False,undef)
			
			displacePlayer :: !Displacements !Footballer -> Footballer
			displacePlayer displacements fb
				= case lookup fb.playerID displacements of
					Just pos				= {fb & pos=pos}
					nothing					= fb
			
			reprimandPlayer :: !ClubName !FootballerID !Reprimand !(![Footballer],![Footballer]) -> (![Footballer],![Footballer])
			reprimandPlayer club1 playerID RedCard (team1,team2)
				= splitAt (nr_players_1 - if (playerID.clubName == club1) 1 0) (uneq1++uneq2)
			where
				(uneq1,_,uneq2)				= break1 (identify_player playerID) (team1++team2)
				nr_players_1				= length team1
			reprimandPlayer _ _ _ teams		= teams			
		
/*	performFootballerActions actions succeededActions match
		performs for each football player in match his succeededAction and moves the ball. 
*/	performFootballerActions :: ![FootballerWithAction] ![FootballerWithAction] !Match -> Match
	performFootballerActions actions succeededActions match=:{theField,theBall,team1,team2,seed,nextRandomP,unittime}
		# (seed,ball,new_players1,new_players2)
						= foldl (flip (performAction succeededActions)) (seed,theBall,team1,team2) actions
		# (ball,seed)	= moveFootball theField (new_players1++new_players2) (ball,seed)
		# match			= { match & team1=new_players1, team2=new_players2, theBall=ball, seed=seed }
		= match
	where	
		performAction :: ![FootballerWithAction] !FootballerWithAction !(!RandomSeed,!FootballState,![Footballer],![Footballer]) 
		                                                             -> (!RandomSeed,!FootballState,![Footballer],![Footballer])
		performAction succeededActions initiative (seed,ball,allPlayers1,allPlayers2)
			| isMember initiative succeededActions				// plan has succeeded
				= performAction` initiative (seed,ball,allPlayers1,allPlayers2)
			| otherwise											// plan has failed
				= (seed,ball,map (failThisPlayerAction initiative) allPlayers1,map (failThisPlayerAction initiative) allPlayers2)
		where
			failThisPlayerAction :: !FootballerWithAction !Footballer -> Footballer
			failThisPlayerAction (idea,playerID) fb
				| identify_player playerID fb
									= {fb & effect=Just (failFootballerAction idea)}
				| otherwise			= fb
			
			performAction` :: !FootballerWithAction !(!RandomSeed,!FootballState,![Footballer],![Footballer]) 
							                      -> (!RandomSeed,!FootballState,![Footballer],![Footballer])
			
			performAction` (Move speed angle,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed1,ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				cur_nose			= fb.nose
				cur_speed			= fb.speed
				skills				= skillsAsList fb
				player_has_ball		= ballIsGainedBy playerID ball
				(p,seed1)			= nextRandomP seed
				feasible_angle		= (fromInt (sign angle)) * (setbetween (abs angle) zero (maxRotateAngle fb))
				new_nose			= cur_nose + feasible_angle
				angleDifficulty 	= angleHowFarFromPi (speed.direction-new_nose)
				angleDifference		= angleHowFarFromAngle speed.direction new_nose
				new_stamina			= alter_stamina ball fb   angleDifficulty angleDifference
				new_health			= alter_health  ball fb p angleDifficulty angleDifference
				healthFat			= getHealthStaminaFactor new_health new_stamina
				new_vel				= healthFat * (setbetween speed.velocity zero (maxVelocity fb angleDifficulty angleDifference))
				new_speed			= {speed & velocity=new_vel}
				new_position`		= move_point (scale_vector (unittime * new_vel) (toRVector new_speed.direction)) fb.pos
				new_position		= point_to_rectangle (zero,{px=theField.flength,py=theField.fwidth}) new_position` 
				new_fb				= {fb & stamina = new_stamina
									      , health  = new_health
									      , speed   = new_speed
									      , pos     = new_position
									      , nose    = new_nose
									      , effect  = Just (Moved new_speed feasible_angle)
									  }
			
		/**	Run has become deprecated.
			Rules for running:
			(1) you can't run through another player
			(2) you can't run faster than maximum velocity for a player (depends on Running and Dribbling skill)
			(3) you can't leave field
			(4) running fast lowers your stamina
			(5) running slow increases your stamina and health
			(6) poor health or poor stamina lowers your maximum velocity
			performAction` (Run speed,playerID) (seed,ball,team1,team2)
				| isEmpty bumpedInto	// no collision with other player
					# new_fb		= {fb & pos     = new_position
									      , speed   = new_speed
									      , stamina = new_stamina
									      , effect  = Just (Ran {event_speed & direction = event_speed.direction - fb.speed.direction})
									  }
					# (team1,team2)	= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
					= (seed2,ball,team1,team2)
				| otherwise				// collission with other player
					# player_speed	= {fb.speed & direction=new_speed.direction, velocity=0.3*new_speed.velocity}
					# new_fb		= {fb & speed   = player_speed
									      , stamina = new_stamina
									      , effect  = Just (Ran {fb.speed & velocity=0.3*new_speed.velocity})
									  }
					# (team1,team2)	= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
					# (p3,seed3)	= nextRandomP seed2
					# (seed4,team1,team2)
									= moveBumpedPlayers {new_speed & velocity=0.3*new_speed.velocity, direction = 0.4*p3*new_speed.direction}
										bumpedInto (seed3,team1,team2)
					= (seed4,ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(p1,seed1)			= nextRandomP seed
				(p2,seed2)			= nextRandomP seed1
				angleDifficulty 	= angleHowFarFromPi speed.direction
				angleDifference		= if (isNothing fb.effect) zero (case fromJust fb.effect of
																	      Ran s	-> angleHowFarFromAngle speed.direction s.direction
																	      _		-> zero)				
				player_has_ball		= ballIsGainedBy playerID ball
				new_stamina			= alter_stamina ball fb    angleDifficulty angleDifference
				new_health			= alter_health  ball fb p1 angleDifficulty angleDifference
				healthFat			= getHealthStaminaFactor new_health new_stamina
				new_vel				= healthFat * (setbetween speed.velocity zero (maxVelocity fb angleDifficulty angleDifference))
				new_speed			= {fb.speed & velocity=new_vel}
				new_angle			= if (p2 == one) (fb.speed.direction + speed.direction)
									 (if (p2 > 0.5)  (fb.speed.direction + (0.85 + 0.15 * healthFat) * pi - pi + speed.direction)
									                 (fb.speed.direction + (1.15 - 0.15 * healthFat) * pi - pi + speed.direction))
				event_speed			= {fb.speed & velocity=new_vel, direction=new_angle}
				new_position`		= move_point (scale_vector (unittime * new_vel) (toRVector new_angle)) fb.pos
				new_position		= point_to_rectangle (zero,{px=theField.flength,py=theField.fwidth}) new_position` 
				bumpedInto			= [fb.playerID \\ fb <- uneq1++uneq2 | inRadiusOfFootballer new_position fb]
		*/	
		/** Rotate has become deprecated:
			Rules for rotating:
			(1) Rotating with slow velocity increases your stamina and health
			(2) Rotating with high velocity lowers your stamina and health
			(3) poor health or poor stamina lowers your maximum velocity
			(4) poor health or poor stamina lowers your precision with turning
			(5) you can't leave field
			(6) you can't run faster than your maximum velocity
			(7) you can't run through another player
			performAction` (Rotate speed,playerID) (seed,ball,team1,team2)
				| not (isEmpty bumpedInto)
					# player_speed			= {fb.speed & direction=new_angle, velocity=0.3*new_speed.velocity}
					# new_fb				= {fb & speed   = player_speed
											      , stamina = new_stamina
											      , effect  = Just (Rotated {direction=new_angle,velocity=0.3*new_speed.velocity})
											  }
					# (team1,team2)			= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
					# (p2,seed2)			= nextRandomP seed1
					# (seed3,team1,team2)	= moveBumpedPlayers {new_speed & velocity=0.3*new_speed.velocity, direction = 0.4*p2*new_speed.direction} 
												bumpedInto (seed2,team1,team2)
					= (seed3,ball,team1,team2)
				| otherwise
					# new_fb				= {fb & pos     = new_position
											      , speed   = new_speed
											      , stamina = new_stamina
											      , effect  = Just (Rotated new_speed)
											  }
					# (new_team1,new_team2)	= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
					= (seed1,ball,new_team1,new_team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(p1,seed1)			= nextRandomP seed
				new_v`				= healthFat * (setbetween speed.velocity zero (maxVelocity fb zero zero))	
				new_v				= if (new_angle >= maxRotateAngle {fb & speed={fb.speed & velocity=new_v`}}) (max (new_v` - 0.5) zero) new_v`
				new_angle			= let lengthFactor	= if (isMember Rotating skills) (fb.length-0.2) fb.length
									      healthFactor	= (healthFat/(0.75*lengthFactor))*speed.direction
									   in if (speed.direction > zero) (min (  maxRotateAngle fb)  healthFactor)
									                                  (max (~(maxRotateAngle fb)) healthFactor)
				skills				= skillsAsList fb
				player_has_ball		= ballIsGainedBy playerID ball
				new_stamina			= alter_stamina ball fb    zero zero
				new_health			= alter_health  ball fb p1 zero zero
				healthFat			= getHealthStaminaFactor new_health new_stamina
				new_direction		= fb.speed.direction + new_angle
				new_speed			= {fb.speed & direction=new_direction, velocity=new_v}
				new_position`		= move_point (scale_vector (unittime * new_v) (toRVector new_direction)) fb.pos
				new_position		= point_to_rectangle (zero,{px=theField.flength,py=theField.fwidth}) new_position` 
				bumpedInto			= [fb.playerID \\ fb <- uneq1++uneq2 | inRadiusOfFootballer new_position fb]
		*/	
		/**	Rules for gaining ball:
			(1) ball obtains position and surface speed of obtaining player
		*/	performAction` (GainBall,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed,GainedBy playerID,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				new_fb				= {fb & effect = Just (GainedBall Success)}
			
		/**	Rules for kicking ball:
			(1) kicking decreases stamina
			(2) kicking is more effective towards your direction, and least effective in opposite direction
			(3) being taller, you can kick harder
			(4) a low stamina/health lower your max kickspeed
			(5) todo: kicking a ball held/gained by a keeper, may damage the keeper
		*/	performAction` (KickBall {vxy={velocity=v,direction=d},vz},playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed2,Free new_ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(p1,seed1)			= nextRandomInt seed
				(p2,seed2)			= nextRandomP   seed1
				new_fb				= {fb & stamina=new_stamina,effect=Just (KickedBall (Just new_speed))}
				theBall				= getFootball ball (team1 ++ team2)
				skills				= skillsAsList fb
				fatHealth			= getHealthStaminaFactor fb.health fb.stamina
				max_v				= maxVelocityBallKick fb
				new_v				= speed_factor * (setbetween v zero max_v)
				new_vz				= speed_factor * (setbetween vz zero max_v)
				new_speed			= {vxy={velocity=new_v,direction=new_direction},vz=new_vz}
				new_stamina			= kickingPenalty fb new_v * fb.stamina
				speed_factor		= oppositeKickPenalty fb d
				new_ball			= {theBall & ballSpeed=new_speed}
				new_direction
					| p2 == one		= d
					# failure		= one - if (isMember Kicking skills) (makeRandomRealisticSkilled p2) (makeRandomRealistic p2)
					| isEven p1
						# new_d		= d - failure * maxKickingDeviation fb
						= if (new_d < zero) (new_d + 2.0*pi) new_d
					| otherwise
						# new_d		= d + failure * maxKickingDeviation fb
						= if (new_d > 2.0*pi) (new_d - 2.0*pi) new_d
			
		/**	Rules for heading ball:
			(1) heading decreases stamina, but less than kicking
			(2) kicking is more effective towards your direction, and least effective in opposite direction
			(3) a low stamina/health lower your max headspeed, but less than kicking
			(4) heading is less harder than kicking, but is not effected by your length
			(5) todo: heading a ball held/gained by a keeper, may damage the keeper (less than with kicking)
		*/	performAction` (HeadBall {vxy={velocity=v,direction=d},vz},playerID) (seed,ballstate,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed2,Free new_ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(p1,seed1)			= nextRandomInt seed
				(p2,seed2)			= nextRandomP   seed1
				skills				= skillsAsList fb
				fatHealth			= getHealthStaminaFactor fb.health fb.stamina
				ball				= getFootball ballstate (team1 ++ team2)
				ball_speed			= ball.ballSpeed.vxy.velocity
				max_v				= maxVelocityBallHead fb ball_speed
				new_v				= setbetween v zero max_v
				new_vz				= 0.25 * (setbetween vz zero max_v)
				new_direction		
					| p2 == one		= d
					# failure		= one - if (isMember Heading skills) makeRandomRealisticSkilled makeRandomRealistic p2
					| isEven p1
						# new_d		= d - failure * maxHeadingDeviation fb
						= if (new_d < zero) (new_d + 2.0*pi) new_d
					| otherwise
						# new_d		= d + failure * maxHeadingDeviation fb
						= if (new_d > 2.0*pi) (new_d - 2.0*pi) new_d
				new_speed			= {vxy={velocity=new_v,direction=new_direction},vz=new_vz}
				new_stamina			= headingPenalty fb new_v ball_speed * fb.stamina
				new_fb				= {fb & stamina=new_stamina,effect=Just (HeadedBall (Just new_speed))}
				new_ball			= {ball & ballSpeed=new_speed}
			
		/**	Rules for feinting:
			(1) you must have velocity in order to feint manouvre.
			(2) a feint manouvre changes your position, and decreases your velocity (depends on Feinting skill)
		*/	performAction` (Feint d,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed,ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				player_has_ball		= ballIsGainedBy playerID ball
				new_stamina			= maxFatigueLossAtFeint fb * fb.stamina
				fatHealth			= getHealthStaminaFactor fb.health fb.stamina
				new_velocity		= fatHealth * fb.speed.velocity * maxVelocityLossAtFeint fb
				new_speed			= {fb.speed & velocity=new_velocity}
				(leftv,rightv)		= orthogonal fb.speed.direction
				sidestep			= case d of FeintLeft -> leftv; _ -> rightv
				new_position`		= move_point ((scale_vector (maxFeintStep fb) (toRVector sidestep))
				                                                 + 
				                                  (scale_vector (unittime * new_velocity) (toRVector fb.speed.direction))
				                                 ) fb.pos
				new_position		= point_to_rectangle (zero,{px=theField.flength,py=theField.fwidth}) new_position`
				new_fb				= {fb & pos=new_position,speed=new_speed,stamina=new_stamina,effect=Just (Feinted d)}
			
		/** Rules for Tackling
			(1) tackling may lower the health of the victim but increases his stamina (last is because he lies on the ground the next rounds)
			(2) tackling costs stamina
		*/	performAction` (Tackle victimID ve,playerID) (seed,ball,team1,team2)
				= (seed1,new_ball,team1T,team2T)
			where
				nrPlayersTeam1		= length team1
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(team1N,team2N)		= splitAt nrPlayersTeam1 (unbreak1 (uneq1,new_fb,uneq2))
				(uneq1T,fbT,uneq2T)	= break1 (identify_player victimID) (team1N ++ team2N)
				(team1T,team2T)		= splitAt nrPlayersTeam1 (unbreak1 (uneq1T,new_target,uneq2T))
				new_stamina_self	= maxFatigueLossAtTackle fb * fb.stamina
				fatHealth_self		= getHealthStaminaFactor fb.health fb.stamina
				new_fb				= {fb & stamina = new_stamina_self, effect = Just (Tackled victimID ve Success)}
				target_has_ball		= ballIsGainedBy victimID ball
				(p,seed1)			= nextRandomP seed
				new_v`				= min max_tackle_velocity ve
				max_tackle_velocity	= 10.0
				new_v				= new_v`/10.0 
				healthDamageTarget	= new_v * fatHealth_self * (0.5*p + 0.1) + (fbT.length-min_length)/2.0
				new_health_target	= max 0.0 fbT.health - healthDamageTarget
				new_target			= {fbT & health = new_health_target, effect = Just (OnTheGround 3) }
				new_ball			= if target_has_ball (Free (mkFootball fbT.pos fbT.speed)) ball	
			
		/** Rules for Schwalbe
			(1) Schwalbe cures stamina
			(2) Performing a Schwalbe when ball was gained causes to lose ball
		*/	performAction` (Schwalbe,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed,ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				new_fb				= {fb & effect = Just (OnTheGround 1)}
				
		/** Rules for catching
			(1) ball optains speed and distance of player
		*/	performAction` (CatchBall,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed,GainedBy playerID,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				new_fb				= {fb & effect=Just (CaughtBall Success)}
			
		/** Rules for playing theater
			(1) playingTheater costs stamina
			(2) Performing a Schwalbe when ball was gained causes to lose ball
		*/	performAction` (PlayTheater,playerID) (seed,ball,team1,team2)
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed,ball,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				effect				= fb.effect
				wasOnTheGround		= if (isNothing effect) False (isOnTheGround (fromJust effect))
				new_event			= if (isOnTheGround (fromJust effect)) (fromJust effect) PlayedTheater
				new_fb				= {fb & effect = Just new_event}
			
		moveBumpedPlayers :: !Speed ![FootballerID] !(!RandomSeed,![Footballer],![Footballer])
												  -> (!RandomSeed,![Footballer],![Footballer])
		moveBumpedPlayers new_speed bumpedInto (seed,team1,team2)
			= foldl (moveBumpedPlayer new_speed) (seed,team1,team2) bumpedInto
		where
			moveBumpedPlayer :: !Speed !(!RandomSeed,![Footballer],![Footballer]) !FootballerID
									 -> (!RandomSeed,![Footballer],![Footballer])
			moveBumpedPlayer new_speed (seed,team1,team2) playerID
				# (team1,team2)		= splitAt (length team1) (unbreak1 (uneq1,new_fb,uneq2))
				= (seed1,team1,team2)
			where
				(uneq1,fb,uneq2)	= break1 (identify_player playerID) (team1 ++ team2)
				(p,seed1)			= nextRandomP seed
				new_pos				= let a = p*2.0*pi in move_point {dx=0.5*cos a,dy=0.5*sin a} fb.pos
				new_fb				= {fb & speed=new_speed, pos=new_pos}
				
	
	/**	moveFootball moves the football one unit, taking into account the surface and air resistance.
	*/	moveFootball :: !FootballField ![Footballer] !(!FootballState,!RandomSeed) -> (!FootballState,!RandomSeed)
		moveFootball _ _ gained=:(GainedBy playerID,seed)
			= gained
		moveFootball field allPlayers (Free football=:{ballSpeed={vxy={velocity=v,direction=d},vz},ballPos},seed)
			= (Free {football & ballSpeed=new_speed,ballPos=new_ballpos},seed1)
		where
			in_the_air				= ballPos.pz > zero
			resistance				= if in_the_air air_resistance surface_resistance
			surface_movement		= scale_vector (unittime * v) (toRVector d)
			new_speed2D				= let new_v = resistance*v in {direction = d, velocity = if (new_v <= 0.05) zero new_v}
//			new_vz`					= if in_the_air (vz - unittime*accelleration_sec) zero
			new_vz`					= vz - unittime*accelleration_sec
			new_height`				= ballPos.pz + vz
			(new_height,new_vz)		= if (new_height` < zero) (0.5*(abs new_height`),let new_v = 0.33*(abs new_vz`) in if (new_v <= 0.8) zero new_v)	// ball bounces, loss of velocity
									                          (new_height`,new_vz`)
			new_speed`				= {vxy=new_speed2D, vz=new_vz}
			new_ballpos				= {pxy=move_point surface_movement ballPos.pxy,pz=new_height}
			field_dimensions		= (zero,{px=field.flength,py=field.fwidth})
			(new_speed,seed1)		= ballBouncesAgainst field new_ballpos new_speed` allPlayers seed
			
			//if the ball bounces against something, velocity will be reduced again and the direction will be changed.
			ballBouncesAgainst :: !FootballField Position3D Speed3D ![Footballer] !RandomSeed -> (!Speed3D,!RandomSeed)
			ballBouncesAgainst field new_ballpos new_speed=:{vxy={velocity=v,direction=d},vz=s3d} allPlayers seed
				//the ball may hit one of the poles of one of the goal or one of the players and bounce away
				//ball hits one of the goal poles
				| againstGoalWestNorthPole || againstGoalWestSouthPole || againstGoalEastNorthPole || againstGoalEastSouthPole
					// 50% bounce left, 50% bounce right
					= ({new_speed & vxy = {new_speed.vxy & direction = if (p1<=0.5) (d-p2*pi) (d+p2*pi), velocity = resistance*v}},seed2)
				//ball hits the top of one of the goals
				| againstGoalWestPoleUpper || againstGoalEastPoleUpper
					# forward_or_back	= if (p1<=0.5) Forward Back
					# up_or_down		= if (new_ballpos.pz < goal_height+(goal_pole_width/2.0)) Down Up
					= bounceBall up_or_down (bounceBall forward_or_back (new_speed,seed2))
				//ball hits one of the footballers				
				| any (\fb -> inRadiusOfFootballer new_ballpos.pxy fb && fb.length >= new_ballpos.pz) allPlayers
					//bounces pure at random (player might get ported)
					= ({new_speed & vxy = {new_speed.vxy & direction = p2*2.0*pi, velocity = resistance*v}, vz=p1*s3d},seed2)
				//ball hits nothing
				| otherwise
					= (new_speed,seed2) 
			where
				(p1,seed1)					= nextRandomP seed
				(p2,seed2)					= nextRandomP seed1
				(northPole,southPole)		= goal_poles field
				againstGoalWestNorthPole 	= inCircleRadiusOfPosition new_ballpos (goal_pole_width/2.0) goal_height {px=0.0,          py=northPole - goal_pole_width/2.0}
				againstGoalWestSouthPole	= inCircleRadiusOfPosition new_ballpos (goal_pole_width/2.0) goal_height {px=0.0,          py=southPole + goal_pole_width/2.0}
				againstGoalEastNorthPole	= inCircleRadiusOfPosition new_ballpos (goal_pole_width/2.0) goal_height {px=field.flength,py=northPole - goal_pole_width/2.0}
				againstGoalEastSouthPole	= inCircleRadiusOfPosition new_ballpos (goal_pole_width/2.0) goal_height {px=field.flength,py=southPole + goal_pole_width/2.0}
				againstGoalWestPoleUpper	= (isbetween new_ballpos.pxy.py (northPole-goal_pole_width/2.0) (southPole+goal_pole_width/2.0))
													&&
											  (isbetween new_ballpos.pz goal_height (goal_height+goal_pole_width))
													&&
											  (new_ballpos.pxy.px <= zero)
				againstGoalEastPoleUpper	= (isbetween new_ballpos.pxy.py (northPole-goal_pole_width/2.0) (southPole+goal_pole_width/2.0))
													&&
											  (isbetween new_ballpos.pz goal_height (goal_height+goal_pole_width))
													&&
											  (new_ballpos.pxy.px >= field.flength)
			
	advanceTime :: !Match -> Match
	advanceTime match=:{playingTime, unittime}
		= {match & playingTime = max zero ((playingTime*160.0 - unittime)/160.0)}


/*	Attribute altering functions depending on angles:
	params: 
		Angle :: between zero and pi, how much the player is running backwards (pi is backwards).
		Angle :: between zero and pi, the difference between the desired angle and the angle the player previously ran to.
*/
alter_stamina :: !FootballState !Footballer !Angle !Angle -> Stamina
alter_stamina ballState fb angleDifficulty angleDifference
	| velocity <= rfv											// increase stamina
		| stamina < MinimumFatigue	= MinimumFatigue
		| otherwise					= stamina^0.8 	
	| otherwise													// lower stamina
		| stamina > MaximumFatigue	= MaximumFatigue * (one - angleDifficulty/(4.0*pi))
		| otherwise					= fv * (one - angleDifficulty/(4.0*pi))
where
	velocity						= fb.speed.velocity
	length							= fb.length
	stamina							= fb.stamina
	rfv								= restore_stamina_velocity ballState fb angleDifficulty angleDifference
	diff							= velocity-rfv
	fv								= if (diff >= 6.0) (stamina^(stamina^(1.6 + 2.0*length/100.0)))
									 (if (diff >= 4.0) (stamina^(1.5 + length/100.0))
									 (if (diff >= 2.0) (stamina^(1.4 - length/100.0))
									                   (stamina^(1.3 - 2.0*length/100.0))))

alter_health :: !FootballState !Footballer !P !Angle !Angle -> Health
alter_health ballState fb p angleDifficulty angleDifference
	| fb.speed.velocity <= rfv		= min (health+p/10.0) 1.0	// increase health
	| otherwise						= health
where
	health							= fb.health
	rfv								= restore_stamina_velocity ballState fb angleDifficulty angleDifference

restore_stamina_velocity :: !FootballState !Footballer !Angle !Angle -> Velocity
restore_stamina_velocity ballState fb angleDifficulty angleDifference
	| ballIsGainedBy fb.playerID ballState
									= max_v / (if (isMember Running   skills) 1.6 2.6)
	| isMember Running skills		= max_v / (if (isMember Dribbling skills) 2.0 3.0) * 1.22 
	| otherwise						= max_v / (if (isMember Dribbling skills) 2.0 3.0)
where
	skills							= skillsAsList fb
	max_v							= maxVelocity fb angleDifficulty angleDifference

maxVelocity :: !Footballer !Angle !Angle -> Velocity
maxVelocity fb angleDifficulty angleDifference
	= dribblingPenalty * runningPenalty * base_velocity
where
	skills							= skillsAsList fb
	base_velocity					= 10.0
	dribblingPenalty				= if (isMember Dribbling skills) 0.95 0.85
	runningPenalty					= if (isMember Running   skills) 1.0  0.85

MinimumFatigue						:== 0.05
MaximumFatigue						:== 0.985


/**	The functions below defines the penalty factor: values between 0.0 and 1.0 that define the loss of an attribute of an action.
*/
::	PenaltyFactor :== Real			// a value between 0.0 and 1.0

kickingPenalty :: !Footballer !Velocity -> PenaltyFactor
kickingPenalty fb new_v				= 1.0 - (if (isMember Kicking (skillsAsList fb)) 0.3 0.6) * (new_v/max_v)^2.0
where
	max_v							= maxVelocityBallKick fb

headingPenalty :: !Footballer !Velocity !Velocity -> PenaltyFactor
headingPenalty fb new_v ball_v		= 1.0 - (if (isMember Heading (skillsAsList fb)) 0.08 0.13) * (new_v/max_v)^2.0
where
	max_v							= maxVelocityBallHead fb ball_v

maxFatigueLossAtTackle :: !Footballer -> PenaltyFactor
maxFatigueLossAtTackle fb			= if (isMember Tackling (skillsAsList fb)) 0.99 0.9

maxFatigueLossAtFeint :: !Footballer -> PenaltyFactor
maxFatigueLossAtFeint fb			= if (isMember Feinting (skillsAsList fb)) 0.92 0.77

maxVelocityLossAtFeint :: !Footballer -> PenaltyFactor
maxVelocityLossAtFeint fb			= if (isMember Feinting (skillsAsList fb)) 0.99 0.75

oppositeKickPenalty :: !Footballer !Angle -> PenaltyFactor
oppositeKickPenalty fb kick_to		= 1.0 - skillPenaltyFactor * (angleHowFarFromPi angle)/pi
where
	angle							= abs (fb.nose - kick_to)
	skills							= skillsAsList fb
	skillPenaltyFactor				= if (isAllMember [Rotating,Kicking] skills) 0.3
									 (if (isAnyMember [Rotating,Kicking] skills) 0.5
									                                             0.9)
