implementation module FootballerFunctions

import Footballer

returnAI					:: FootballerAction -> FootballerAI m
returnAI action				= \{BrainInput | memory} -> (action,memory)

halt						:: FootballerAI m
halt						= returnAI (Move zero zero)

rotate						:: Angle -> FootballerAI m
rotate angle				= returnAI (Move zero angle)

ahead						:: Velocity -> FootballerAI m
ahead v						= \{me,memory} -> (Move {direction=me.nose,velocity=v} zero,memory)

fix							:: Position Metre -> FootballerAI m
fix point eps				= fix`
where
	fix` input=:{me}
		| distance <= eps	= halt     input
		| otherwise			= returnAI (Move {direction=angle,velocity=v} (angle - me.nose)) input
	where
		distance			= dist            me.pos point
		angle				= angleWithObject me.pos point
		v					= max 6.0 distance

kick						:: Position -> FootballerAI m
kick point					= kick`
where
	kick` input=:{football,others,me}
		| dist me.pos ball.ballPos <= maxKickReach me
							= returnAI (KickBall {vxy = {direction=angle,velocity=v},vz=1.0}) input
		| otherwise			= halt input
	where
		ball				= getBall input
		angle				= angleWithObject me.pos point
		v					= 2.0*dist me.pos point

track_ball					:: FootballerAI m
track_ball					= track_ball`
where
	track_ball` input		= fix (getBall input).ballPos.pxy (maxKickReach input.me) input

centerOfGoal				:: !Home !FootballField -> Position
centerOfGoal home field		= {py = field.fwidth/2.0, px = if (home==West) zero field.flength}

(team) infix 9				:: !Footballer ![Footballer] -> [Footballer]
(team) player players		= filter (sameClub player) players

(opponents) infix 9			:: !Footballer ![Footballer] -> [Footballer]
(opponents) player players	= filter (not o (sameClub player)) players

getBall						:: !(BrainInput m) -> Football
getBall input				= getFootball input.football [input.me : input.others]

northwest					:: FootballField -> Position
northwest field				= zero

southwest					:: FootballField -> Position
southwest field				= {px=0.0,py=field.fwidth}

northeast					:: FootballField -> Position
northeast field				= {px=field.flength,py=0.0}

southeast					:: FootballField -> Position
southeast field				= {px=field.flength,py=field.fwidth}
